/*
 * procsms.c - part of procsms application
 *
 * Copyright (C) 2003-2006 BLADOX, s.r.o.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 */

#include <config.h>
#include <turbo/turbo.h>

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "procsms.h"
#include "archive.h"
#include "crypted.h"
#include "masked.h"

/*
TODO:

- optimize sms_rec - reread only modified(=masked or crypted)
- APDU rec_nr not direct
- ?mask all
- ?better view crypted (show head)
*/

/* *INDENT-OFF* */

static lc_char PROGMEM lc_Enter_Text[]={
        LC_EN("Enter Text:")
        LC_CZ("Text zpravy:")
	LC_FR("Entrez Texte:")
	LC_DE("Text der Meldung")
	LC_DK("Skriv Tekst:")
        LC_END
};


static lc_char PROGMEM lc_SIM_SMS_Full[]={
#if defined CONFIG_MASKED && defined CONFIG_CRYPTED
        LC_EN("SIM SMS Memory Full! Crypted and Masked Messages Cannot Be Received.")
        LC_CZ("SIM SMS pamet je plna! Nelze prijimat sifrovane ci maskovane zpravy.")
	LC_FR("M\5moire SIM SMS pleine!. Les messages crypt\5s ou masqu\5s ne seront pas recus!.")
LC_DE("SIM SMS Speicher ist voll! Empfangen von verschl\176sselten Nachrichten nich m\174glich")
#else
#if defined CONFIG_MASKED
        LC_EN("SIM SMS Memory Full! Masked Messages Cannot Be Received.")
        LC_CZ("SIM SMS pamet je plna! Nelze prijimat maskovane zpravy.")
#else
        LC_EN("SIM SMS Memory Full! Secure Messages Cannot Be Received.")
        LC_CZ("SIM SMS pamet je plna! Nelze prijimat sifrovane zpravy.")
	LC_DE("SIM SMS Speicher ist voll! Empfangen von verschl\176sselten Nachrichten nich m\174glich")
	LC_FR("M\5moire SIM SMS pleine!. Les messages crypt\5s ne seront pas recus!.")
	LC_DK("SIM SMS Memory er fyldt! Krypteret SMS kan ikke modtages.")
#endif
#endif
        LC_END
};

/*
TODO
static lc_char PROGMEM lc_Send[]={
        LC_EN("Send?")
        LC_CZ("Poslat?")
	LC_FR("Envoyer?")
	LC_DE("Schicken?")
	LC_DK("Send?")
        LC_END
};
*/

lc_char PROGMEM lc_Msg_Reply[]={
        LC_EN("Reply")
        LC_CZ("Odpovedet")
	LC_FR("R\5pondre")
	LC_DE("Antworten")
	LC_DK("Besvar")
        LC_END
};

lc_char PROGMEM lc_Msg_Forward[]={
        LC_EN("Forward")
        LC_CZ("Preposlat")
	LC_FR("En avant")
	LC_DE("Weiter schicken")
	LC_DK("Videresend")
        LC_END
};

static lc_char PROGMEM lc_Sending_SMS[]={
        LC_EN("Sending Message")
        LC_CZ("Posilam zpravu")
	LC_FR("Message en cours")
	LC_DE("Sende Nachricht")
	LC_DK("Sender besked")
        LC_END
};

static lc_char PROGMEM lc_SMS_Send_OK[]={
        LC_EN("Message Sent")
        LC_CZ("Zprava odeslana")
	LC_FR("Message envoy\5")
	LC_DE("Nachricht gesendet")
	LC_DK("Besked sendt")
        LC_END
};

static lc_char PROGMEM lc_SMS_Send_Not_OK[]={
        LC_EN("Sending was not successful")
        LC_CZ("Zpravu se nepodarilo odeslat")
	LC_FR("Envoi en \5chec")
	LC_DE("Nachricht konnte nicht gesendet werden")
	LC_DK("Kunne ikke sende besked!")
        LC_END
};

static lc_char PROGMEM lc_SN_Error[]={
        LC_EN("Application locked - wrong serial number.")
        LC_CZ("Apliakce zamcena - spatne seriove cislo.")
//TODO	LC_FR("Entrez Texte:")
//TODO	LC_DE("Text der Meldung")
//TODO	LC_DK("Skriv Tekst:")
        LC_END
};


u8 PROGMEM t_Menu_Alpha[] = "Secure";

#ifdef CONFIG_ONDISPLAY
static lc_char PROGMEM lc_On_Display[] = {
  	LC_EN("Flash SMS") 
	LC_CZ("Na displej")
	LC_FR("Affichage actif")
	LC_DE("Direktanzeige")
	LC_DK("Flash Besked")
	LC_END
};
#endif

/* *INDENT-ON* */

#ifdef CONFIG_CRYPTED
//static u8 PROGMEM def_main_key[] = { 'a', 'd', 'g', 'j', 'm', 'p', 't', 'w' };
static u8 PROGMEM def_main_key[] = { '1', '2', '3', '4' };
static u8 PROGMEM crypted_head[] = { 0x45, 0xe7, 0x50, 0x9a, 0x85, 0x52, 0x8b, 0x44, 0x97, 0xcb, 0xd5, 0x28 };	//="ENCRYPTED...\r" in 7b" (0xcb -> version)
static u8 PROGMEM t_Kill_mime[] = "kill";

u8 reset_ok;
#endif

#ifdef CONFIG_MASKED
static u8 PROGMEM t_Default_Mask_Sender[] = { "1189" };
#endif

u8 PROGMEM ef_sms_path[] = { 0x3f, 0x00, 0x7f, 0x10, 0x6f, 0x3c };

Smsing_mem *glob_mem;

void insert_sorted_alpha (SEdge * e, SEdge ** start, u8 * alpha)
{
//adn sorted
  u8 i, len;
  u8 *t;
  SEdge *f, *g;
  SNode *n;

  dbsp (">insert_sorted_alpha\n");

  g = f = rw (start);
  while (f)
  {
    dbsp ("ADN insert sorted\n");
    n = rw (&f->t);
    t = rw (&n->text);
    len = strlen (t);
    dbsp ("L ");
    dbch (len);
    dbc (' ');
    dbch (alpha[1]);
    dbc ('\n');
    if (len > alpha[0])
      len = alpha[0];
    for (i = 0; i < len; i++)
    {
      dbch (alpha[1 + i]);
      dbc (' ');
      dbch (rb (t + i));
      dbc ('\n');
      if (tolower (alpha[1 + i]) > tolower (rb (t + i)))
	goto X_NEXT;
    }
    dbsp ("X_BREAK\n");
    break;
  X_NEXT:
    dbsp ("X_NEXT\n");
    g = f;
    f = rw (&f->next);
  }

  if (f == g)
  {
    dbsp ("F==G\n");
    ww (start, e);
  }
  else
  {
    dbsp ("F!=G\n");
    ww (&g->next, e);
  }

  ww (&e->next, f);
  dbsp ("<insert_sorted_alpha\n");
}

u8 check_crypted (u8 * s)	// return version 1,3, or 0=not crypted
{
  u8 i;
  u8 *t;

  t = tpdu_seek (s, T_SMS_UDL);
//  if (*t != 0xA0)
//    return 0;

  t++;
  /*
     for (i = 0; i < sizeof (crypted_head); i++)
     if (rb (crypted_head + i) != t[i])
     return 0;
   */

  if (memcmp (t, crypted_head, sizeof (crypted_head)) == 0)
    return 1;			// version 1
  if (memcmp (t, crypted_head, 9) == 0)
    return t[9];		// version 3...
  return 0;
}

u8 check_mask (u8 * s, u8 p1)
{
#if defined CONFIG_MASKED
  // p1==0xff -> archive
  Mask_key *key = glob_mem->mask_keys;
  u8 *t;
  SEdge *msisdns;
  SNode *n;
  u8 i;
  Crypted_msg *msg;

  t = tpdu_seek (s, T_SMS_AA);
  while (key)
  {
    if (glob_mem->mask_locked == 1 || key != glob_mem->masked_key)
    {
      msisdns = rw (&key->mask_msisdns);
      while (msisdns)
      {
	n = rw (&msisdns->t);
	if (msisdncmp (t, MSISDN_SMS, rw (&n->p), MSISDN_ADN))
	{
	  u8 nr_masks = rb (&key->nr_masks);

	  dbsp ("MASK_FOUND\n");

	  if (p1 != 0xFF)
	  {
	    // delete from crypted list

	    msisdns = glob_mem->crypted_msgs;
	    while (msisdns)
	    {
	      n = msisdns->t;
	      msg = n->p;
	      if (msg->p1 == p1)
		goto X_HAS_CRYPTED;
	      msisdns = msisdns->next;
	    }
	    goto X_HAS_NOT_CRYPTED;

	  X_HAS_CRYPTED:
	    {
	      SEdge *msisdns = glob_mem->crypted_msgs;
	      SEdge *e;
	      SNode *n;
	      Crypted_msg *msg;

	      e = msisdns;
	      while (msisdns)
	      {
		n = msisdns->t;
		msg = n->p;
		if (msg->p1 == p1)
		{
		  dbsp ("delete crypted msg\n");
		  free (msg);
//XXXMS           free (msg->msisdn);
		  free (n->text);
		  free (n);

		  // delete crypted msg
		  if (e == msisdns)
		    glob_mem->crypted_msgs = e->next;
		  else
		    e->next = msisdns->next;

		  free (msisdns);
		  break;
		}

		e = msisdns;
		msisdns = msisdns->next;
	      }

	    }

	  X_HAS_NOT_CRYPTED:
	  }

	  if (nr_masks == 0)
	    substitute_mask (s, locale (lc_Default_Mask),
			     t_Default_Mask_Sender);
	  else
	  {
	    SEdge *m_e = rw (&key->mask_masks);
	    SNode *m_n;
	    Mask *mask;

	    i = p1;
	    if (nr_masks < i)
	      i = i % nr_masks;
	    else
	      i = nr_masks % i;

	    dbsp ("RND_MASK ");
	    dbch (nr_masks);
	    dbc (' ');
	    dbch (p1);
	    dbc (' ');
	    dbch (i);
	    dbc (' ');
	    dbc ('\n');

	    while (i)
	    {
	      m_e = rw (&m_e->next);
	      i--;
	    }
	    m_n = rw (&m_e->t);
	    mask = rw (&m_n->p);

	    substitute_mask (s, rw (&mask->text), rw (&mask->sender));	// mask->text MEM_SLV
	  }
	  goto X1;
	}
	msisdns = rw (&msisdns->next);
      }
    }

    key = rw (&key->next);
  }
  return 0;
X1:
  return 1;
#else
  return 0;
#endif
}

void reset_me (void *x)
{
  dbsp (">reset_me\n");
  reset_ok = 1;
  refresh (Q_REFRESH_SIM_RESET, 0, 0, 0);
  dbsp ("reset_me_1\n");
  while (1);
  dbsp ("<reset_me\n");
}

void warn_full (void *x)
{
  u8 len = 0;
  u8 *buf = buf_A ();
  u8 *text;
  u8 i;

  dbsp (">warn_full\n");
  if (glob_mem->warn == 1)
    return;
  glob_mem->warn = 1;

  text = locale (lc_SIM_SMS_Full);

  play_tone (text, TONE_POSITIVE, DURATION_SEC, 1);
//  display_text_raw (text, Q_DISPLAY_TEXT_DELAY_CLEAR);

/*
  len = strlen (text);

  buf[0] = T_TEXT_STRING | CR_FLAG;
  buf[1] = len + 1;
  buf[2] = 0x04;
  memcpy (&buf[3], text, len);
  len += 3;

  stk_cmd (STK_CMD_SET_UP_IDLE_MODE_TEXT, 0x00, DEV_ID_ME, len);
*/

  dbsp ("<warn_full\n");
}

u8 check_free ()		// 1 = FULL, 0 = FREE SMS EF
{
  u8 i, nr_free, nr_unknown;

// count nr free, unknown sms
  nr_free = 0;
  nr_unknown = 0;
  for (i = 0; i < glob_mem->sms_rec_nr; i++)
  {
    if (glob_mem->sms_rec[i] == 0xFF)
      nr_unknown++;
    if ((glob_mem->sms_rec[i] & 0x01) == 0)
      nr_free++;
  }

  dbsp ("check_free ");
  dbch (glob_mem->sms_rec_nr);
  dbc (' ');
  dbch (nr_unknown);
  dbc (' ');
  dbch (nr_free);
  dbc ('\n');

  if (nr_unknown == 0 && nr_free == 0)
    return 1;

  return 0;
}

void create_crypted (u8 p1, u8 * s)
{
  u8 *ms;
  u8 *ms_nr;
  SEdge *f, *g;
  Crypted_msg *m;
  SNode *n;
  u8 i;
  u8 *t;
  SEdge *msisdns;
  Crypted_msg *msg;

  msisdns = glob_mem->crypted_msgs;

  dbsp ("create crypted item\n");

  n = malloc (sizeof (SNode));
  if (n == NULL)
    goto X_END_2;

  msisdns = malloc (sizeof (SEdge));
  if (msisdns == NULL)
  {
    free (n);
    goto X_END_2;
  }

  msg = malloc (sizeof (Crypted_msg));
  if (msg == NULL)
  {
    free (n);
    free (msisdns);
    goto X_END_2;
  }

  t = tpdu_seek (s, T_SMS_AA);

  for (i = 0; i < 8; i++)
    msg->msisdn[i] = 0;
  memcpy (msg->msisdn, t, *t / 2 + 3);
  ms = get_adn (msg->msisdn, MSISDN_SMS);

#ifdef DEBUG
  {
    u8 ik;
    u8 il = strlen (ms);

    dbsp ("MS_X ");
    for (ik = 0; ik < il; ik++)
      dbc (ms[ik]);
    dbc ('\n');
  }
#endif

  if (ms == NULL)
  {
    free (msisdns);
    free (n);
    free (msg);
    goto X_END_2;
  }

  msg->p1 = p1;
  msg->adn_ready = adn_ready ();
  t = tpdu_seek (s, T_SMS_SCTS);
  memcpy (&msg->time, t, 6);

  n->text = ms;
  n->cb = crypted_msg_ctx;
  n->p = msg;

  msisdns->f = &crypted_n;
  msisdns->t = n;
  msisdns->source = MEM_F_P | MEM_T_R;

  //insert time sorted

  g = f = glob_mem->crypted_msgs;
  while (f)
  {
    dbsp ("insert time sorted\n");

    n = f->t;
    m = n->p;
    for (i = 0; i < 6; i++)
    {
      dbsp ("C: ");
      dbch (swap_nibble (msg->time[i]));
      dbc (' ');
      dbch (swap_nibble (m->time[i]));
      dbc ('\n');
      if (swap_nibble (msg->time[i]) > swap_nibble (m->time[i]))
	goto X_UP_PREV;
      if (swap_nibble (msg->time[i]) < swap_nibble (m->time[i]))
	break;
    }
    g = f;
    f = f->next;
  }
X_UP_PREV:
  if (f == g)
    glob_mem->crypted_msgs = msisdns;
  else
    g->next = msisdns;

  msisdns->next = f;

X_END_2:
}

void delete_crypted (u8 p1)
{
  SEdge *msisdns = glob_mem->crypted_msgs;
  SEdge *e;
  SNode *n;
  Crypted_msg *msg;

  e = msisdns;
  while (msisdns)
  {
    n = msisdns->t;
    msg = n->p;
    if (msg->p1 == p1)
    {
      dbsp ("delete crypted msg ");
      dbih (n);
      dbc ('\n');
      free (msg);
      free (n->text);
      free (n);

      // delete crypted msg
      if (e == msisdns)
	glob_mem->crypted_msgs = e->next;
      else
	e->next = msisdns->next;

      free (msisdns);
      break;
    }

    e = msisdns;
    msisdns = msisdns->next;
  }
  glob_mem->sms_rec[p1 - 1] = 0x00;
}

void fake_sms_file (File_apdu_data * fa)
{
  u8 i;

  dbsp ("FAKE_SMS_FILE\n");
  db_malloc ();

  if (fa->ins == ME_CMD_GET_RESPONSE)
  {
    u16 nr_sms;
    u8 i;

    dbsp ("FAKE_GET_RESPONSE\n");
    sim (fa->ins, fa->p1, fa->p2, fa->p3, fa->data);
    nr_sms = ((fa->data[2] << 8) | fa->data[3]) / fa->data[14];
    dbsp ("#SMS ");
    dbih (nr_sms);
    dbc ('\n');

    if (nr_sms < glob_mem->sms_rec_nr)
      glob_mem->sms_rec_nr = nr_sms;

    nr_sms = (u16) glob_mem->sms_rec_nr * fa->data[14];
    fa->data[2] = nr_sms >> 8;
    fa->data[3] = nr_sms;

    return;
  }

  if (fa->ins == ME_CMD_READ_RECORD)
  {
    SEdge *msisdns;
    SNode *n;
    u8 *t, *s;
    u8 i, len;
    Crypt_key *c_key;
    Crypted_msg *msg;
    Pers_mem *pers = app_data ();

    dbsp ("FAKE_READ_RECORD\n");
    sim (fa->ins, fa->p1, fa->p2, fa->p3, fa->data);

    s = fa->data;
    if (fa->p1 > glob_mem->sms_rec_nr)
      goto X1;

    glob_mem->sms_rec[fa->p1 - 1] = *s;
    dbsp ("P ");
    dbch (glob_mem->sms_rec[fa->p1 - 1]);
    dbc ('\n');

    if ((*s & 0x01) == 0x00 || (*s & 0x04) != 0x00)
      goto X1;			// deleted SMS or ME originating SMS

    s++;
    i = *s;
    s = s + 1 + i;		//skip SMSC

    if (check_mask (s, fa->p1))
      goto X1;

    // Not masked

    dbsp ("CR_0\n");
    if (check_crypted (s) == 0)
      goto X1;

    dbsp ("CR_2\n");

    msisdns = glob_mem->crypted_msgs;
    while (msisdns)
    {
      n = msisdns->t;
      msg = n->p;
      if (msg->p1 == fa->p1)
	goto X2;
      msisdns = msisdns->next;
    }

    dbsp ("CREATE CRYPTED READ\n");
    create_crypted (fa->p1, s);

  X2:
    t = tpdu_seek (s, T_SMS_UD);

    if (rw (&pers->crypted_label))
    {
      u8 *label = rw (&pers->crypted_label);

      i = strlen (label);
      memcpy (t, label, i);
    }
    else
    {
      i = strlen (locale (lc_Default_Crypted));
      memcpy (t, locale (lc_Default_Crypted), i);
    }
    dcs_78 (t, i, DCS_8_TO_7);
    *(t - 1) = i;

  X1:
    if (check_free ())
      stk_thread (warn_full, NULL);

    db_malloc ();
    return;
  }

  if (fa->ins == ME_CMD_UPDATE_RECORD)
  {
    u8 *t;
    u8 *s = fa->data;
    u8 i = *s;
    Pers_mem *pers = app_data ();

    glob_mem->warn = 0;

    s = fa->data;

    dbsp ("FAKE_UPDATE_RECORD\n");

    if (fa->p1 > glob_mem->sms_rec_nr)
      goto X_UPDATE_END;

    i = *s;

    dbsp ("I ");
    dbch (i);
    dbc ('\n');
    dbsp ("P ");
    dbch (glob_mem->sms_rec[fa->p1 - 1]);
    dbc ('\n');

    if (glob_mem->sms_rec[fa->p1 - 1] != i && i != 0x03
	&& glob_mem->sms_rec[fa->p1 - 1] != 0x00 && i != 0)
    {
      dbsp ("FAKE_UPDATE_RECORD_READ\n");
      glob_mem->sms_rec[fa->p1 - 1] = *s;
      sim (ME_CMD_READ_RECORD, fa->p1, fa->p2, fa->p3, fa->data);
      *s = glob_mem->sms_rec[fa->p1 - 1];
    }

    if (i == 0x00)
      delete_crypted (fa->p1);

    glob_mem->sms_rec[fa->p1 - 1] = i;

    if (i == 0x03 || i == 0x01)	// received, to be read or read
    {
      s++;
      i = *s;
      s = s + 1 + i;		//skip SMSC

      if (check_crypted (s))
      {
	SEdge *msisdns;
	SNode *n;
	Crypted_msg *msg;

	msisdns = glob_mem->crypted_msgs;

	msisdns = glob_mem->crypted_msgs;
	while (msisdns)
	{
	  n = msisdns->t;
	  msg = n->p;
	  if (msg->p1 == fa->p1)
	    goto X_UPDATE_END_2;
	  msisdns = msisdns->next;
	}

	t = tpdu_seek (s, T_SMS_DCS);
	*t = 0xF2;		// subsitute DCS to 7b

	dbsp ("CREATE CRYPTED UPDATE\n");
	create_crypted (fa->p1, s);
      }
    }

  X_UPDATE_END_2:
    if (check_free ())
      stk_thread (warn_full, NULL);

  X_UPDATE_END:

#ifdef CONFIG_ARCHIVE
    archive_store_sms (fa->data, SMS_PLAIN, fa->p1, fa->p2);
#else
    sim (fa->ins, fa->p1, fa->p2, fa->p3, fa->data);
#endif

    fa->data[0] = 0x90;
    fa->data[1] = 0x00;
    db_malloc ();
    return;
  }

  dbsp ("FAKE_OTHER_CMD\n");
  sim (fa->ins, fa->p1, fa->p2, fa->p3, fa->data);
  db_malloc ();
  return;
}

u8 send_msg (u8 type, u8 * def_text, u8 * def_ms, u8 ucs2)	// def_ms is MSISDN_SMS
{
  u8 *res;
  u8 *ms = NULL;
  u8 *text = def_text;
  u8 i;
  u8 len;
  u8 err;
  u8 dcs;			//0xF6; //0x12; //0xF2;              // Class 2 - SIM specific message, 7b default alphabet
  u8 pid = 0x00;
  u8 *buf = buf_B ();
  u8 *tmp;
  u8 sms_len;
  u8 coding;

  if (glob_mem->c2_dcs_7b == 1)
    dcs = 0xF2;
  else
    dcs = 0xF6;

  if (type == MSG_ON_DISPLAY)
    dcs = 0xF0;			// 7b default alphabet, class 0
  else if (type == MSG_KILL)
  {
    dcs = tsms_dcs ();
    pid = tsms_pid ();
  }

  if (type == MSG_KILL)
    coding = Q_GET_INPUT_ALPHABET;
  else
  {
    if (ucs2)
      coding = Q_GET_INPUT_ALPHABET | Q_GET_INPUT_UCS2;
    else
      coding = Q_GET_INPUT_ALPHABET | Q_GET_INPUT_SMS_PACKED;
  }

X_TEXT:
  if (type == MSG_CRYPTED)
  {
    if (ucs2)
      len = 59;			// TODO ucs2 ???
    else
      len = 136;
  }
  else
  {
    if (ucs2)
      len = 80;			// TODO ucs2 ???
    else
      len = 160;
  }

  res = get_input (locale (lc_Enter_Text), 0, len, text, coding);	//text STRING_SLV

  if (res == ENULL)
  {
    i = APP_END;
    goto X_END;
  }

  if (res == NULL)
  {
    i = APP_BACK;
    goto X_END;
  }

  len = res[0] - 1;

// @ fix for broken phones
  if (ucs2 == 0)
  {
    u16 bits = len * 8;

    if (bits != 0 && bits % 7 == 0)
    {
      dbsp ("POSSIBLE @\n");
      if ((res[1 + len] & 0xfe) == 0)	//end @
      {
	dbsp ("POSSIBLE_1 @\n");
	res[1 + len] |= ('\r' << 1);	// @->CR
      }
    }
  }

  if (len > MAX_MSG_LEN - 3)
    len = MAX_MSG_LEN - 3;

  dbsp ("INPUT LEN ");
  dbch (len);
  dbc ('\n');

  text = glob_mem->msg_buf;

  glob_mem->msg_buf[0] = STRING_SLV;
  glob_mem->msg_buf[1] = res[1];
  if (ucs2 == 0)
    glob_mem->msg_buf[2] = (len * 8) / 7;
  else
    glob_mem->msg_buf[2] = len;
  memcpy (&glob_mem->msg_buf[3], res + 2, len);

X_MSISDN:
  if (def_ms == NULL)
  {
    res = msisdn (NULL);

    if (res == ENULL)
    {
      i = APP_END;
      goto X_END;
    }

    if (res == NULL)
    {
      i = APP_BACK;
      goto X_TEXT;
    }

    def_ms = msisdn2str (res, MSISDN_ADN, MEM_R);
    if (def_ms)
    {
      ms = str2msisdn (def_ms, MSISDN_SMS, MEM_R);
      free (def_ms);
      def_ms = NULL;
    }
    else
      ms = NULL;

    if (ms == NULL)
    {
      err = ERR_NO_RAM;
      goto X_ERR;
    }
  }
  else
    ms = def_ms;		//reply

X_KEY:
  if (type == MSG_CRYPTED)
  {
    u8 *out = buf;
    u8 *key = out + 16;
    u8 j;

    tmp = key + 16;

    if (make_main_key () != NO_ERROR)
    {
      err = ERR_NO_RAM;
      goto X_ERR;
    }

    if (glob_mem->unlocked)
    {
      SEdge *msisdns;
      SNode *n;
      Crypt_key *c_key;

      msisdns = glob_mem->crypted_keys;

      while (msisdns)
      {
	n = rw (&msisdns->t);
	c_key = rw (&n->p);
	if (msisdncmp (ms, MSISDN_SMS, rw (&c_key->msisdn), MSISDN_ADN))
	{
	  twofish_ecb_decrypt (&c_key->text, key, &glob_mem->ak);
	  goto X1;
	}
	msisdns = rw (&msisdns->next);
      }
    }
    //ask for key to key[]

    i = password (key, 16, NULL);

    if (i == APP_BACK)
    {
      if (ms != def_ms)
      {
	free (ms);
	ms = NULL;
      }

      if (def_ms == NULL)
	goto X_MSISDN;
      else
	goto X_TEXT;
    }

    if (i != APP_OK)
      goto X_END;

  X1:
    xrand (out);

#ifdef DEBUG
    dbsp ("RAND: ");
    for (i = 0; i < 8; i++)
    {
      dbch (out[i]);
      dbc (' ');
    }
    dbc ('\n');
#endif

    err = twofish_setup (key, 16, &glob_mem->ak);
    glob_mem->unlock_key_valid = 0;
    if (err != NO_ERROR)
      goto X_ERR;

    memcpy (tmp, crypted_head, sizeof (crypted_head));
    tmp[9] = 3;			// msg version

// 12 bytes = head "ENCRYPTED...\r" 
// 8 bytes = rand, bit 1 of [19] 7b/usc2 (1==ucs2)
// 1 byte = len
// 119 bytes = text 136 7b chars 
// sum: 12+128

    for (i = 0; i < 8; i++)
      tmp[12 + i] = out[i];	//rand

    if (ucs2)
      tmp[19] |= 0x01;		//ucs2
    else
      tmp[19] &= 0xfe;		//no ucs2

    tmp[20] = glob_mem->msg_buf[2];

    for (i = 0; i < len; i++)
      tmp[21 + i] = glob_mem->msg_buf[3 + i];
    for (; i < 119; i++)
      tmp[21 + i] = 140 - len;	//padding

    err = twofish_cbc_encrypt (&tmp[12], &tmp[12], 8, &glob_mem->ak);
    if (err != NO_ERROR)
      goto X_ERR;

    sms_len = 140;
  }
  else if (type == MSG_KILL)
  {
    dbsp ("SEND KILL\n");
    tmp = buf;
    res = create_head_msg (tmp, t_Kill_mime);
    *res = len;
    res++;
    memcpy (res, &glob_mem->msg_buf[3], len);
    sms_len = res - tmp + len + 1;	// head+1b_len+msg_len
  }
  else
  {
    memcpy (buf, &glob_mem->msg_buf[3], len);
    tmp = buf;
    sms_len = len;
  }

X_SEND:

  dbsp (">SEND_SMS DCS ");
  dbch (dcs);
  dbc ('\n');
  res =
    send_sms (tmp, sms_len, ms, MSISDN_SMS | SMS_SRR, dcs, pid,
	      locale (lc_Sending_SMS), NULL);
  dbsp ("<SEND_SMS\n");

  i = get_tag (res, T_RESULT);
  if (i != 0)
  {
    i += 2;
    if (res[i] == 0x00)
    {
      i = display_text (locale (lc_SMS_Send_OK), NULL);
      if (type == MSG_CRYPTED && dcs == 0xF6)
	glob_mem->c2_dcs_7b = 0;
    }
    else
    {
      if (type == MSG_KILL && pid != 0x00)
      {
	dbsp ("RESEND_SMS PID 0x00\n");
	pid = 0x00;
	goto X_SEND;
      }
      if (type == MSG_CRYPTED && dcs != 0xF6)
      {
	dbsp ("RESEND_SMS 0xF6\n");
	dcs = 0xF6;
	goto X_SEND;
      }
      i = display_text (locale (lc_SMS_Send_Not_OK), NULL);
    }

    if (i == APP_OK)
    {
      i = APP_BACK;
      goto X_END;
    }

    if (i == APP_END)
      goto X_END;

    if (ms != def_ms)
    {
      free (ms);
      ms = NULL;
    }

    if (def_ms == NULL)
      goto X_MSISDN;
    else
      goto X_TEXT;
  }

  i = APP_BACK;
  goto X_END;

X_ERR:
  i = APP_OK;

X_END:
  if (ms != def_ms)
    free (ms);

  if (i == APP_OK)
    return perror (err);
  return i;
}

u8 check_sn ();

void display_wrong_sn ()
{
  display_text (locale (lc_SN_Error), NULL);
}

void smsing (Menu_selection_data * x)
{
  SCtx *c;
  u8 nr_items = 0;
  u8 it_masked = 0;
  u8 it_crypted = 0;
  u8 it_ondisplay = 0;
  u8 it_archive = 0;
  u8 item = x->item;

/*
Order:
- crypted
- masked
- ondisplay
- archive
*/

  if (check_sn () == 0)
  {
    display_wrong_sn ();
    return;
  }

  dbsp ("ITEM :");
  dbch (item);
  dbc ('\n');

#ifdef CONFIG_CRYPTED
  it_crypted = nr_items;
#endif

#ifdef CONFIG_MASKED
  nr_items++;
  it_masked = nr_items;
#endif

#ifdef CONFIG_ONDISPLAY
  nr_items++;
  it_ondisplay = nr_items;
#endif

#ifdef CONFIG_ARCHIVE
  nr_items++;
  it_archive = nr_items;
#endif

  if (item == it_ondisplay)
  {
#ifdef CONFIG_ONDISPLAY
    dbsp ("menu_ondisplay\n");
    send_msg (MSG_ON_DISPLAY, NULL, NULL, 0);
    return;
#endif
  }

  c = spider_init ();
  if (c == NULL)
    return;
  spider_set_order (c, ORDER_P, ORDER_R, ORDER_E, ORDER_D);

  if (item == it_crypted)
  {
#ifdef CONFIG_CRYPTED
    // crypted
    dbsp ("menu_crypted\n");
    if (check_free ())
    {
      glob_mem->warn = 0;
      warn_full (NULL);
    }

#if 0
// update MSISDN->ADN, needed for phones not doing this automatically, e.g.7610
    if (adn_ready ())
    {
      u8 i = 0;
      SEdge *msisdns = glob_mem->crypted_msgs;
      SNode *n;
      Crypted_msg *msg;

      while (msisdns)
      {
	n = msisdns->t;
	msg = n->p;
	if (msg->adn_ready == 0)
	{
	  i = 1;
	  break;
	}
	msisdns = msisdns->next;
      }

      if (i)
      {
	u8 path[6];

// TODO no refresh but manual read??? - znam msisdn, muzu u adn_unknown nahradit primo

	dbsp (">refresh sms\n");
	memcpy (path, ef_sms_path, 6);
	refresh (Q_REFRESH_SIM_FILE_CHANGE, 1, 6, path);
	dbsp ("<refresh sms\n");
      }
    }
#endif

    c->n = &crypted_n;
    c->eP = &crypted_edges_p;
#endif
  }
  else if (item == it_masked)
  {
#ifdef CONFIG_MASKED
    // masked
    dbsp ("menu_masked\n");
    c->n = &masked_n;
    c->eP = &masked_edges_p;
#endif
  }
  else if (item == it_archive)
  {
#ifdef CONFIG_ARCHIVE
    // archive
    dbsp ("menu_archive\n");
    spider_set_order (c, ORDER_R, ORDER_P, ORDER_E, ORDER_D);
    seek_sms ();

    c->n = &archive_n;
    c->eP = &archive_edges_p;
    c->df_fnc = archive_df;
    glob_mem->is_rules = 0;
    glob_mem->drop = 0;
#endif
  }

  spider (c);

#ifdef CONFIG_CRYPTED
  glob_mem->unlocked = 0;
#endif

#ifdef CONFIG_MASKED
  free (glob_mem->masked_new_key);
  glob_mem->masked_new_key = NULL;
#endif
}

void turbo_handler (u8 action, void *data)
{
  switch (action)
  {
    case ACTION_APP_REGISTER:
      {
	Pers_mem *p = emalloc (sizeof (Pers_mem));
	u8 i;
	u8 *buf = malloc (3 * 16);
	u8 *in = buf;
	u8 *tmp = in + 16;
	u8 *key = tmp + 16;
	Twofish_key *ak = malloc (sizeof (Twofish_key));

        set_proc_8 (PROC_8_CONFIG_SETUP, 0);
	set_proc_8 (PROC_8_CONFIG_SIM_FIRST, 0);
	set_proc_8 (PROC_8_CONFIG_BINARY_CHOICE, 0);

#ifdef CONFIG_CRYPTED
	reg_sms_tag (locale (lc_RESET_SMS), t_Kill_mime, SMS_TYPE_MSG);
#endif

#ifdef CONFIG_ARCHIVE
	reg_file_type (t_ext_sms, file_sms_cb);
#endif

	if (p && buf)
	{
#ifdef CONFIG_MASKED
	  ww (&p->mask_keys, NULL);
#endif

#ifdef CONFIG_ARCHIVE
	  wb (&p->activ, 0);
#endif

#ifdef CONFIG_CRYPTED
	  ww (&p->crypted_keys, NULL);
	  ww (&p->reset_sms, NULL);
	  ww (&p->crypted_label, NULL);
	  memcpy (tmp, def_main_key, sizeof (def_main_key));
	  make_crypt_key (key, tmp, sizeof (def_main_key));
#ifdef DEBUG
	  dbsp ("DEF_KEY: ");
	  for (i = 0; i < 16; i++)
	  {
	    dbch (key[i]);
	    dbc (' ');
	  }
	  dbc ('\n');
#endif
	  xrand (in);
	  xrand (in + 8);
	  memcpy (&p->main_key_rnd, in, 16);
	  twofish_setup (key, 16, ak);
	  twofish_ecb_encrypt (in, &p->main_key_crnd, ak);
#ifdef DEBUG
	  dbsp ("KEY_RND: ");
	  for (i = 0; i < 16; i++)
	  {
	    dbch (rb (&p->main_key_rnd[i]));
	    dbc (' ');
	  }
	  dbc ('\n');
	  dbsp ("KEY_CRND: ");
	  for (i = 0; i < 16; i++)
	  {
	    dbch (rb (&p->main_key_crnd[i]));
	    dbc (' ');
	  }
	  dbc ('\n');

#endif
#endif
	}
	reg_app_data (p);

	free (buf);
	free (ak);
      }
      break;
    case ACTION_APP_UNREGISTER:
      {
	Pers_mem *p = app_data ();
	SEdge *e, *f;
	SNode *n;
	u8 *t;

#ifdef CONFIG_CRYPTED
	Crypted_msg *msg;
	Crypt_key *ckey;
#endif

#ifdef CONFIG_MASKED
	Mask *mask;
	Mask_key *m, *m1;
#endif

	dbsp ("UNREG crypted_keys\n");

#ifdef CONFIG_CRYPTED
	efree (rw (&p->reset_sms));
	efree (rw (&p->crypted_label));
	e = rw (&p->crypted_keys);
	while (e)
	{
	  n = rw (&e->t);
	  t = rw (&n->text);
	  ckey = rw (&n->p);
	  efree (rw (&ckey->msisdn));
	  efree (ckey);
	  efree (t);
	  efree (n);
	  f = rw (&e->next);
	  efree (e);
	  e = f;
	}
#endif

#ifdef CONFIG_MASKED
	dbsp ("UNREG mask_keys\n");
	m = rw (&p->mask_keys);
	while (m)
	{
	  dbsp ("UNREG mask_masks\n");
	  e = rw (&m->mask_masks);
	  while (e)
	  {
	    n = rw (&e->t);
	    t = rw (&n->text);
	    mask = rw (&n->p);
	    efree (rw (&mask->sender));
	    efree (rw (&mask->text));
	    efree (mask);
	    efree (t);
	    efree (n);
	    f = rw (&e->next);
	    efree (e);
	    e = f;
	  }
	  dbsp ("UNREG mask_msisdns\n");
	  e = rw (&m->mask_msisdns);
	  while (e)
	  {
	    n = rw (&e->t);
	    t = rw (&n->text);
	    efree (rw (&n->p));
	    efree (t);
	    efree (n);
	    f = rw (&e->next);
	    efree (e);
	    e = f;
	  }

	  m1 = rw (&m->next);
	  efree (m);
	  m = m1;
	}
#endif

	dbsp ("UNREG p\n");
	efree (p);

	if (glob_mem)
	{
	  free (glob_mem->sms_rec);
#ifdef CONFIG_CRYPTED
	  dbsp ("UNREG crypted_msgs\n");
	  e = glob_mem->crypted_msgs;
	  while (e)
	  {
	    n = e->t;
	    t = n->text;
	    msg = n->p;
	    free (msg);
//XXXMS free(msg->msisdn);
	    free (t);
	    free (n);
	    f = e->next;
	    free (e);
	    e = f;
	  }
#endif
	  free (glob_mem);
	}
	dbsp ("UNREG_DONE\n");
      }
      break;
    case ACTION_APP_INIT:
      {
	Pers_mem *p = app_data ();
	u8 i;

	dbsp ("PROCSMS INIT\n");
	reg_file (ef_sms_path, 3);

	glob_mem = malloc (sizeof (Smsing_mem));
	if (glob_mem)
	{
	  dbsp ("PROCSMS INIT_1 ");
	  dbih (sizeof (Smsing_mem));
	  dbc (' ');
	  dbih (sizeof (Crypted_msg));
	  dbc ('\n');
#ifdef CONFIG_MASKED
	  glob_mem->mask_locked = 1;
	  glob_mem->mask_keys = rw (&p->mask_keys);
	  glob_mem->masked_new_key = NULL;
#endif

#ifdef CONFIG_CRYPTED
	  glob_mem->unlock_key_valid = 0;
	  glob_mem->unlocked = 0;
	  glob_mem->crypted_keys = rw (&p->crypted_keys);
	  glob_mem->crypted_msgs = NULL;
	  glob_mem->c2_dcs_7b = 1;
#endif

#ifdef CONFIG_ARCHIVE
	  glob_mem->activ = rb (&p->activ);
#endif

	  glob_mem->sms_rec = malloc (DEF_SMS_REC_NR);
	  glob_mem->sms_rec_nr = DEF_SMS_REC_NR;
	  if (glob_mem->sms_rec)
	    for (i = 0; i < DEF_SMS_REC_NR; i++)
	      glob_mem->sms_rec[i] = 0xFF;
	  glob_mem->warn = 0;
	}

      }
      break;
    case ACTION_INSERT_MENU:
      set_menu_alpha (t_Menu_Alpha);
#ifdef CONFIG_CRYPTED
      insert_menu (locale (lc_Crypted));
#endif

#ifdef CONFIG_MASKED
      insert_menu (locale (lc_Masked));
#endif

#ifdef CONFIG_ONDISPLAY
      insert_menu (locale (lc_On_Display));
#endif

#ifdef CONFIG_ARCHIVE
      if (mc_detect () == MC_OK)
	insert_menu (locale (lc_Archive));
#endif
      break;
    case ACTION_MENU_SELECTION:
      stk_thread (smsing, data);
      break;
#ifdef CONFIG_CRYPTED
    case ACTION_SMS_MSG:
      {
	u8 *t = skip_head (data);
	u8 l = *t;
	Pers_mem *pers = app_data ();
	u8 *label = rw (&pers->reset_sms);

	dbsp ("REC KILL\n");
	if (label && l == strlen (label))
	{
	  u8 i;

	  t++;
	  for (i = 0; i < l; i++)
	    if (t[i] != rb (label + i))
	    {
	      dbsp ("NOT KILL\n");
	      return;
	    }
	  reset_ok = 0;
	  stk_thread (reset_me, NULL);
	  dbsp ("KILL_1\n");
	  if (reset_ok == 0)
	    while (1);
	  dbsp ("KILL_2\n");
	}
      }
      break;
#endif
    case ACTION_FILE_APDU:
      fake_sms_file (data);
      break;
    default:
      break;
  }
}
